@@ -92,9 +92,12 @@ def parse(text):
     return {"text": text, "html": ToParse(text, "html"), "tex": ToParse(text, "tex")}
 
 
-def parse_define(
-    directive, rest, substitutions, substitutions_match, substitution_groups
-):
+def parse_define(directive, rest, defines):
+    defines["substitutions"] = defines.get("substitutions", {})
+    defines["substitutions_match"] = defines.get("substitutions_match", [])
+    defines["substitution_groups"] = defines.get("substitution_groups", [])
+    defines["substitution_ranges"] = defines.get("substitution_ranges", {})
+
     if directive == "MATCH":
         regex = r"\[(.*)\]\s+\[(.*)\]"
         matches = re.match(regex, rest)
@@ -107,7 +110,7 @@ def parse_define(
             raise SyntaxError(
                 "DEFINE MATCH must have at least as many replacements as it has directives"
             )
-        substitutions_match.append(
+        defines["substitutions_match"].append(
             {"directives": directives_list, "replacements": replacements_list}
         )
     elif directive == "GROUP":
@@ -122,14 +125,27 @@ def parse_define(
             blocks[i] = tuple(block[1:-1].split(" "))
         if not all(len(block) == len(blocks[0]) for block in blocks):
             raise SyntaxError("DEFINE GROUP blocks must all be of the same length")
-        substitution_groups.append(
+        defines["substitution_groups"].append(
             {
                 "directives": blocks[0],
-                "replacements": list_to_dict([list_to_dict(block) for block in blocks]),
+                "replacements": list_to_dict(
+                    [list_to_dict(block) for block in blocks[1:]]
+                ),
             }
         )
+    elif directive == "RANGE":
+        blocks = rest.split(" ")
+        if len(blocks) != 3:
+            raise SyntaxError("DEFINE RANGE takes exactly three arguments")
+        directive, low, high = blocks
+        try:
+            low = int(low)
+            high = int(high)
+        except ValueError:
+            raise SyntaxError("DEFINE RANGE bounds must be integers")
+        defines["substitution_ranges"][directive] = [low, high]
     else:
-        substitutions[directive] = rest.split(" ")
+        defines["substitutions"][directive] = rest.split(" ")
 
 
 def parse_input_lines(lines):
@@ -218,18 +234,17 @@ def consume_rest_of_solution(buff, end):
 def consume_rest_of_question(buff, id_factory):
     contents = []
     input_lines = []
-    substitutions = {}
-    substitutions_match = []
-    substitution_groups = []
+    defines = {}
     solution = None
     solution_note = None
     config = {}
+    template = []
     while True:
         line = buff.pop()
         mode, directive, rest = parse_directive(line)
         if mode is None:
             if input_lines and line.strip():
-                raise SyntaxError("Unexpected content in QUESTION after INPUT")
+                template.append(line)
             elif not input_lines:
                 contents.append(line)
         elif mode == "INPUT":
@@ -245,13 +260,30 @@ def consume_rest_of_question(buff, id_factory):
                 )
         elif mode == "END":
             if directive == "QUESTION":
-                question_type, options, option_solutions = parse_input_lines(
-                    input_lines
-                )
+                (
+                    question_type,
+                    options,
+                    option_solutions,
+                ) = parse_input_lines(input_lines)
 
                 if option_solutions and solution:
                     raise SyntaxError("Received multiple solutions.")
 
+                template = "\n".join(template)
+                if template:
+                    if question_type in ("long_answer", "long_code_answer"):
+                        # OK
+                        pass
+                    elif question_type in ("short_answer", "short_code_answer"):
+                        if "\n" in template:
+                            raise SyntaxError(
+                                "Cannot have newlines in template for INPUT SHORT ANSWER"
+                            )
+                    else:
+                        raise SyntaxError(
+                            f"Cannot have a template for question type {question_type}"
+                        )
+
                 return {
                     "id": id_factory.get_id(config.get("ID")),
                     "type": question_type,
@@ -263,22 +295,15 @@ def consume_rest_of_question(buff, id_factory):
                     **parse("\n".join(contents)),
                     "config": config,
                     "options": options,
-                    "substitutions": substitutions,
-                    "substitutions_match": substitutions_match,
-                    "substitution_groups": substitution_groups,
+                    "template": template,
+                    **defines,
                 }
             else:
                 raise SyntaxError(
                     f"Unexpected END {directive if directive else line} in QUESTION"
                 )
         elif mode == "DEFINE":
-            parse_define(
-                directive,
-                rest,
-                substitutions,
-                substitutions_match,
-                substitution_groups,
-            )
+            parse_define(directive, rest, defines)
         elif mode == "CONFIG":
             config[directive] = rest
         else:
@@ -291,9 +316,7 @@ def consume_rest_of_group(buff, end, id_factory):
     group_contents = []
     elements = []
     started_elements = False
-    substitutions = {}
-    substitutions_match = []
-    substitution_groups = []
+    defines = {}
     pick_some = None
     scramble = False
     inline = False
@@ -331,17 +354,13 @@ def consume_rest_of_group(buff, end, id_factory):
                 "type": "group",
                 **parse("\n".join(group_contents)),
                 "elements": elements,
-                "substitutions": substitutions,
-                "substitutions_match": substitutions_match,
-                "substitution_groups": substitution_groups,
+                **defines,
                 "pick_some": pick_some,
                 "scramble": scramble,
                 "inline": inline,
             }
         elif mode == "DEFINE":
-            parse_define(
-                directive, rest, substitutions, substitutions_match, substitution_groups
-            )
+            parse_define(directive, rest, defines)
         elif mode == "CONFIG":
             if directive == "PICK":
                 if pick_some:
@@ -366,9 +385,8 @@ def _convert(text, *, path=None, allow_random_ids=True):
     groups = []
     public = None
     config = {}
-    substitutions = {}
-    substitutions_match = []
-    substitution_groups = []
+    defines = {}
+    watermark = None
     if path is not None:
         buff = load_imports(text, path)
     else:
@@ -387,6 +405,8 @@ def _convert(text, *, path=None, allow_random_ids=True):
                     "SCRAMBLE_OPTIONS",
                 ]:
                     config[directive.lower()] = [int(x) for x in rest.split(" ") if x]
+                elif directive == "WATERMARK":
+                    watermark = dict(brightness=int(rest))
                 else:
                     raise SyntaxError(
                         "Unexpected CONFIG directive {}".format(directive)
@@ -410,13 +430,7 @@ def _convert(text, *, path=None, allow_random_ids=True):
                 else:
                     groups.append(group)
             elif mode == "DEFINE":
-                parse_define(
-                    directive,
-                    rest,
-                    substitutions,
-                    substitutions_match,
-                    substitution_groups,
-                )
+                parse_define(directive, rest, defines)
             else:
                 raise SyntaxError(f"Unexpected directive: {line}")
     except SyntaxError as e:
@@ -429,9 +443,8 @@ def _convert(text, *, path=None, allow_random_ids=True):
         "public": public,
         "groups": groups,
         "config": config,
-        "substitutions": substitutions,
-        "substitutions_match": substitutions_match,
-        "substitution_groups": substitution_groups,
+        **defines,
+        "watermark": watermark,
         "version": VERSION,
     }
 
